# -*- coding: utf-8 -*-
"""mesh_label_predicting.ipynb

Automatically generated by Colaboratory.

Original file is located at
    https://colab.research.google.com/drive/1iuKgfU9aR3aLZ_W-Kpkipy_1vOa4DtI0
"""

# install all dependencies and imports 
!pip install --upgrade Tensorflow
!pip install tensorflow_graphics

import glob
import os
import numpy as np
import pickle

import tensorflow.compat.v1 as tf
tf.disable_v2_behavior()
tf.enable_eager_execution() #old version of tensorflow then need to enbale earger execution

from tensorflow_graphics.nn.layer import graph_convolution as graph_conv
from tensorflow_graphics.notebooks import mesh_segmentation_dataio as dataio
from tensorflow_graphics.notebooks import mesh_viewer


import tensorflow_graphics.geometry.transformation as tfg_transformation
from tensorflow_graphics.notebooks import threejs_visualization

# mount google drive for data file access
# Load the Drive helper and mount
from google.colab import drive
from google.colab import files

# This will prompt for authorization.
drive.mount('/content/drive')

drive.mount("/content/drive", force_remount=True)

#defining functions

def mesh_encoder(batch_mesh_data, num_filters, output_dim, conv_layer_dims):
  """A mesh encoder using feature steered graph convolutions.

    The shorthands used below are
      `B`: Batch size.
      `V`: The maximum number of vertices over all meshes in the batch.
      `D`: The number of dimensions of input vertex features, D=3 if vertex
        positions are used as features.

  Args:
    batch_mesh_data: A mesh_data dict with following keys
      'vertices': A [B, V, D] `float32` tensor of vertex features, possibly
        0-padded.
      'neighbors': A [B, V, V] `float32` sparse tensor of edge weights.
      'num_vertices': A [B] `int32` tensor of number of vertices per mesh.
    num_filters: The number of weight matrices to be used in feature steered
      graph conv.
    output_dim: A dimension of output per vertex features.
    conv_layer_dims: A list of dimensions used in graph convolution layers.

  Returns:
    vertex_features: A [B, V, output_dim] `float32` tensor of per vertex
      features.
  """
  batch_vertices = batch_mesh_data['vertices']

  # Linear: N x D --> N x 16.
  vertex_features = tf.keras.layers.Conv1D(16, 1, name='lin16')(batch_vertices)

  # graph convolution layers
  for dim in conv_layer_dims:
    with tf.variable_scope('conv_%d' % dim):
      vertex_features = graph_conv.feature_steered_convolution_layer(
          vertex_features,
          batch_mesh_data['neighbors'],
          batch_mesh_data['num_vertices'],
          num_weight_matrices=num_filters,
          num_output_channels=dim)
    vertex_features = tf.nn.relu(vertex_features)

  # Linear: N x 128 --> N x 256.
  vertex_features = tf.keras.layers.Conv1D(
      256, 1, name='lin256')(
          vertex_features)
  vertex_features = tf.nn.relu(vertex_features)

  # Linear: N x 256 --> N x output_dim.
  vertex_features = tf.keras.layers.Conv1D(
      output_dim, 1, name='lin_output')(
          vertex_features)

  return vertex_features


def get_learning_rate(params):
  """Returns a decaying learning rate."""
  global_step = tf.train.get_or_create_global_step()
  learning_rate = tf.train.exponential_decay(
      params['init_learning_rate'],
      global_step,
      params['lr_decay_steps'],
      params['lr_decay_rate'])
  return learning_rate


def model_fn(features, labels, mode, params):
  """Returns a mesh segmentation model_fn for use with tf.Estimator."""
  logits = mesh_encoder(features, params['num_filters'], params['num_classes'],
                        params['encoder_filter_dims'])
  predictions = tf.argmax(logits, axis=-1, output_type=tf.int32)
  outputs = {
      'vertices': features['vertices'],
      'triangles': features['triangles'],
      'num_vertices': features['num_vertices'],
      'num_triangles': features['num_triangles'],
      'predictions': predictions,
  }
  # For predictions, return the outputs.
  if mode == tf.estimator.ModeKeys.PREDICT:
    outputs['labels'] = features['labels']
    return tf.estimator.EstimatorSpec(mode=mode, predictions=outputs)
  # Loss
  # Weight the losses by masking out padded vertices/labels.
  vertex_ragged_sizes = features['num_vertices']
  mask = tf.sequence_mask(vertex_ragged_sizes, tf.shape(labels)[-1])
  loss_weights = tf.cast(mask, dtype=tf.float32)
  loss = tf.losses.sparse_softmax_cross_entropy(
      logits=logits, labels=labels, weights=loss_weights)
  # For training, build the optimizer.
  if mode == tf.estimator.ModeKeys.TRAIN:
    optimizer = tf.train.AdamOptimizer(
        learning_rate=get_learning_rate(params),
        beta1=params['beta'],
        epsilon=params['adam_epsilon'])
    update_ops = tf.get_collection(tf.GraphKeys.UPDATE_OPS)
    with tf.control_dependencies(update_ops):
      train_op = optimizer.minimize(
          loss=loss, global_step=tf.train.get_global_step())
    return tf.estimator.EstimatorSpec(mode=mode, loss=loss, train_op=train_op)

  # For eval, return eval metrics.
  eval_ops = {
      'mean_loss':
          tf.metrics.mean(loss),
      'accuracy':
          tf.metrics.accuracy(
              labels=labels, predictions=predictions, weights=loss_weights)
  }
  return tf.estimator.EstimatorSpec(
      mode=mode, loss=loss, eval_metric_ops=eval_ops)

#setting up parameters

MODEL_PARAMS = {
    'num_filters': 5,
    'num_classes': 6,
    'encoder_filter_dims': [16, 32, 64],
}

train_data_files = glob.glob("/content/drive/My Drive/meshes/tfrecord_folder/consistent_aligned_R-mixed-train_meshes/*train*.tfrecord*")

test_data_files = glob.glob("/content/drive/My Drive/meshes/tfrecord_folder/consistent_aligned_R-mixed-train_meshes/*test*.tfrecord*")

retrain_model_dir = '/content/drive/My Drive/meshes/tfrecord_folder/consistent_aligned_R-mixed-train_meshes/tfrecord_retrain/'

train_io_params = {
    'batch_size': 5,
    'parallel_threads': 5,
    'is_training': True,
    'shuffle': True,
    'sloppy': True,
}

eval_io_params = {
    'batch_size': 5,
    'parallel_threads': 5,
    'is_training': False,
    'shuffle': False
}


def train_fn():
  return dataio.create_input_from_dataset(dataio.create_dataset_from_tfrecords,
                                          train_data_files, train_io_params)


def eval_fn():
  return dataio.create_input_from_dataset(dataio.create_dataset_from_tfrecords,
                                          test_data_files, eval_io_params)


train_params = {
    'beta': 0.9,
    'adam_epsilon': 1e-8,
    'init_learning_rate': 0.001,
    'lr_decay_steps': 10000,
    'lr_decay_rate': 0.95,
}

train_data_files #view training data files

test_data_files #view testing data files

retrain_model_dir # view save folder

"""Check that Run time type is set to GPU - else the training takes ages and is super slow."""

# Run training model from scratch using TFRecord set

train_params.update(MODEL_PARAMS)

checkpoint_delay = 120  # Checkpoint every 2 minutes.
max_steps = 100000  # Number of training steps.

config = tf.estimator.RunConfig(
    log_step_count_steps=1,
    save_checkpoints_secs=checkpoint_delay,
    keep_checkpoint_max=3)

classifier = tf.estimator.Estimator(
    model_fn=model_fn,
    model_dir=retrain_model_dir,
    config=config,
    params=train_params)

train_spec = tf.estimator.TrainSpec(input_fn=train_fn, max_steps=max_steps)

eval_spec = tf.estimator.EvalSpec(
    input_fn=eval_fn,
    steps=None,
    start_delay_secs=2 * checkpoint_delay,
    throttle_secs=checkpoint_delay)

print('Start training & eval.')
tf.estimator.train_and_evaluate(classifier, train_spec, eval_spec)
print('Train and eval done.')

#Test model

test_io_params = {
    'is_training': False,
    'sloppy': False,
    'shuffle': True,
    'repeat': False
}
test_tfrecords = test_data_files

def predict_fn():
  return dataio.create_input_from_dataset(dataio.create_dataset_from_tfrecords,
                                          test_tfrecords,
                                          test_io_params)


estimator = tf.estimator.Estimator(model_fn=model_fn,
                                   model_dir=retrain_model_dir,
                                   params=MODEL_PARAMS)
test_predictions = estimator.predict(input_fn=predict_fn)

#visualise results
prediction = next(test_predictions)
input_mesh_data = {
    'vertices': prediction['vertices'],
    'faces': prediction['triangles'],
}
predicted_mesh_data = {
    'vertices': prediction['vertices'],
    'faces': prediction['triangles'],
    'vertex_colors': mesh_viewer.SEGMENTATION_COLORMAP[prediction['predictions']],
}

input_viewer = mesh_viewer.Viewer(input_mesh_data)
prediction_viewer = mesh_viewer.Viewer(predicted_mesh_data)

prediction # show prediction data

np.where(prediction['predictions']>1)

mesh_viewer.SEGMENTATION_COLORMAP[prediction['predictions']]

predicted_mesh_data # show prediction mesh data

np.where(predicted_mesh_data['vertex_colors'] != predicted_mesh_data['vertex_colors'][0])

test_data_files

test_predictions

FileName = os.path.basename(test_data_files[0])
wFileName = os.path.dirname(test_data_files[0]) + "/output/predicted_"+ os.path.splitext(FileName)[0][:-4] + "aligned.dat"
wFileNameMesh = os.path.dirname(test_data_files[0]) + "/output/predicted_" + os.path.splitext(FileName)[0][:-4] + "aligned_mesh_data.dat"

print(wFileName)
print(wFileNameMesh)

#write out prediction files to drive

with open(wFileName, mode='wb') as fl:
    pickle.dump(prediction, fl,protocol = 2)

with open(wFileNameMesh, mode='wb') as fl:
    pickle.dump(predicted_mesh_data, fl, protocol = 2)

print('finished writing out files...')